/**
  *  \file t_ptrmultilist.cc
  *  \brief Test Suite for ptrmultilist.cc
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <stdio.h>
#include "cpluslib/ptrmultilist.h"

#include "tests.h"

namespace {
    void nullDeleter(void* /*p*/)
    { }

    class Obj {
        Obj(const Obj&);
        Obj& operator=(const Obj&);
     public:
        int n;
        static int live;
        Obj(int n)
            : n(n) { ++live; }
        ~Obj()
            { --live; }
    };
    int Obj::live = 0;

    bool objCompare(Obj* a, Obj* b)
        {
            return a->n < b->n;
        }
}

/** Test basic operations for ptr_multi_list_base. */
void
TestPtrMultilist::testBasic()
{
    // Four elements
    int a = 1, b = 2, c = 3, d = 4;

    // A list with a null deleter. Add our variables.
    ptr_multi_list_base ml(nullDeleter);
    TS_ASSERT(ml.empty());
    TS_ASSERT_EQUALS(ml.size(), 0);
    ml.push_back(&a);
    ml.push_back(&b);
    ml.push_back(&c);
    TS_ASSERT(!ml.empty());
    TS_ASSERT_EQUALS(ml.size(), 3);

    // Regular iteration
    {
        ptr_multi_list_base::iterator i = ml.begin();
        TS_ASSERT(i != ml.end());
        TS_ASSERT(i == ml.begin());
        TS_ASSERT(i.get() == &a);
        i.next();
        TS_ASSERT(i != ml.end());
        TS_ASSERT(i != ml.begin());
        TS_ASSERT(i.get() == &b);
        i.next();
        TS_ASSERT(i != ml.end());
        TS_ASSERT(i != ml.begin());
        TS_ASSERT(i.get() == &c);
        i.next();
        TS_ASSERT(i == ml.end());
        TS_ASSERT(i != ml.begin());
    }

    // Regular iteration with two iterators
    {
        ptr_multi_list_base::iterator i = ml.begin();
        ptr_multi_list_base::iterator j = ml.begin();
        TS_ASSERT(i == j);
        TS_ASSERT(i.get() == &a);
        TS_ASSERT(j.get() == &a);
        i.next();
        TS_ASSERT(i != j);
        TS_ASSERT(i.get() == &b);
        TS_ASSERT(j.get() == &a);
        j.next();
        TS_ASSERT(i == j);
        TS_ASSERT(i.get() == &b);
        TS_ASSERT(j.get() == &b);
    }

    // Now, add one item at the front, and iterate again
    ml.push_front(&d);
    {
        ptr_multi_list_base::iterator i = ml.begin();
        TS_ASSERT(i != ml.end());
        TS_ASSERT(i == ml.begin());
        TS_ASSERT(i.get() == &d);
        i.next();
        TS_ASSERT(i != ml.end());
        TS_ASSERT(i != ml.begin());
        TS_ASSERT(i.get() == &a);
        i.next();
        TS_ASSERT(i != ml.end());
        TS_ASSERT(i != ml.begin());
        TS_ASSERT(i.get() == &b);
        i.next();
        TS_ASSERT(i != ml.end());
        TS_ASSERT(i != ml.begin());
        TS_ASSERT(i.get() == &c);
        i.next();
        TS_ASSERT(i == ml.end());
        TS_ASSERT(i != ml.begin());
    }

    // Extract
    void* p = ml.extract_front();
    TS_ASSERT(p == &d);
    p = ml.extract_front();
    TS_ASSERT(p == &a);
    p = ml.extract_front();
    TS_ASSERT(p == &b);
    p = ml.extract_front();
    TS_ASSERT(p == &c);
    TS_ASSERT(ml.empty());
}

/** Test erase for ptr_multi_list_base with multiple iterators. */
void
TestPtrMultilist::testBasicErase1()
{
    int a = 1, b = 2, c = 3;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);
    ml.push_back(&c);

    // Deletion of second element with both iterators pointing there
    {
        ptr_multi_list_base::iterator i = ml.begin();
        ptr_multi_list_base::iterator j = ml.begin();
        i.next();
        j.next();
        ml.erase(i);
        TS_ASSERT(i.get() == 0);
        TS_ASSERT(j.get() == 0);
        i.next();
        j.next();
        TS_ASSERT(i.get() == &c);
        TS_ASSERT(j.get() == &c);
        TS_ASSERT(i == j);
    }

    // List now has two elements. Try deleting the first.
    {
        ptr_multi_list_base::iterator i = ml.begin();
        ptr_multi_list_base::iterator j = ml.begin();
        ml.erase(i);
        TS_ASSERT(i.get() == 0);
        TS_ASSERT(j.get() == 0);
        i.next();
        j.next();
        TS_ASSERT(i.get() == &c);
        TS_ASSERT(j.get() == &c);
    }

    // List now has one element. Must be c.
    {
        ptr_multi_list_base::iterator i = ml.begin();
        TS_ASSERT(i.get() == &c);
    }
}

/** Test erase for ptr_multi_list_base with multiple iterators. */
void
TestPtrMultilist::testBasicErase2()
{
    int a = 1, b = 2, c = 3;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);
    ml.push_back(&c);

    // Deletion of second element with one iterator pointing there
    {
        ptr_multi_list_base::iterator i = ml.begin();
        ptr_multi_list_base::iterator j = ml.begin();
        i.next();
        ml.erase(i);
        TS_ASSERT(i.get() == 0);
        TS_ASSERT(j.get() == &a);
        i.next();
        j.next();
        TS_ASSERT(i.get() == &c);
        TS_ASSERT(j.get() == &c);
        TS_ASSERT(i == j);
    }

    // List now has two elements. Delete first with iterator pointing at second.
    {
        ptr_multi_list_base::iterator i = ml.begin();
        ptr_multi_list_base::iterator j = ml.begin();
        i.next();
        ml.erase(j);
        TS_ASSERT(i.get() == &c);
        TS_ASSERT(j.get() == 0);
        j.next();
        TS_ASSERT(i.get() == &c);
        TS_ASSERT(j.get() == &c);
        TS_ASSERT(i == j);
    }
}

/** Test erase for ptr_multi_list_base with multiple iterators. */
void
TestPtrMultilist::testBasicErase3()
{
    int a = 1, b = 2;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);

    // List has two elements. Delete second with iterator pointing at first.
    {
        ptr_multi_list_base::iterator i = ml.begin();
        ptr_multi_list_base::iterator j = ml.begin();
        i.next();
        ml.erase(i);
        TS_ASSERT(i.get() == 0);
        TS_ASSERT(j.get() == &a);
        i.next();
        j.next();
        TS_ASSERT(i.get() == 0);
        TS_ASSERT(j.get() == 0);
        TS_ASSERT(i == j);
    }
}

/** Test multiple iterators of nontrivial lifetime. */
void
TestPtrMultilist::testBasicMultiIter()
{
    int a = 1, b = 2, c = 3;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);
    ml.push_back(&c);

    // Make three iterators pointing at a, b, c, respectively.
    ptr_multi_list_base::iterator* pa = new ptr_multi_list_base::iterator(ml.begin());
    ptr_multi_list_base::iterator* pb = new ptr_multi_list_base::iterator(ml.begin());
    pb->next();
    ptr_multi_list_base::iterator* pc = new ptr_multi_list_base::iterator(ml.begin());
    pc->next();
    pc->next();

    // Validate
    TS_ASSERT(pa->get() == &a);
    TS_ASSERT(pb->get() == &b);
    TS_ASSERT(pc->get() == &c);

    // Delete in random order
    delete pb;
    TS_ASSERT(pa->get() == &a);
    TS_ASSERT(pc->get() == &c);

    delete pa;
    TS_ASSERT(pc->get() == &c);

    delete pc;
}

/** Test ptr_multi_list_base::insert, middle of list. */
void
TestPtrMultilist::testBasicInsert1()
{
    int a = 1, b = 2, c = 3;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);

    // Make an iterator pointing at b
    ptr_multi_list_base::iterator pa = ml.begin();
    pa.next();
    TS_ASSERT(pa.get() == &b);

    // Insert c
    ml.insert(pa, &c);

    // Validate iterator
    TS_ASSERT(pa.get() == &b);
    pa.next();
    TS_ASSERT(pa == ml.end());

    // Validate list
    // (This intentionally uses the assignment operator to test that as well.)
    pa = ml.begin();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &a);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &c);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &b);
    pa.next();
    TS_ASSERT(pa == ml.end());
}

/** Test ptr_multi_list_base::insert, beginning of list. */
void
TestPtrMultilist::testBasicInsert2()
{
    int a = 1, b = 2, c = 3;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);

    // Make an iterator pointing at a
    ptr_multi_list_base::iterator pa = ml.begin();
    TS_ASSERT(pa.get() == &a);
    TS_ASSERT(pa == ml.begin());

    // Insert c
    ml.insert(pa, &c);

    // Validate iterator
    TS_ASSERT(pa != ml.begin());
    TS_ASSERT(pa.get() == &a);

    // Validate list
    pa = ml.begin();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &c);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &a);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &b);
    pa.next();
    TS_ASSERT(pa == ml.end());
}

/** Test ptr_multi_list_base::insert, end of list. */
void
TestPtrMultilist::testBasicInsert3()
{
    int a = 1, b = 2, c = 3;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);

    // Make an iterator pointing after b
    ptr_multi_list_base::iterator pa = ml.begin();
    pa.next();
    pa.next();
    TS_ASSERT(pa == ml.end());

    // Insert c
    ml.insert(pa, &c);
    
    // Validate iterator
    TS_ASSERT(pa == ml.end());

    // Validate list
    pa = ml.begin();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &a);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &b);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &c);
    pa.next();
    TS_ASSERT(pa == ml.end());
}

/** Test ptr_multi_list_base::insert, at deleted element. */
void
TestPtrMultilist::testBasicInsert4()
{
    int a = 1, b = 2, c = 3, d = 4;
    ptr_multi_list_base ml(nullDeleter);
    ml.push_back(&a);
    ml.push_back(&b);
    ml.push_back(&c);

    // Make an iterator pointing at b
    ptr_multi_list_base::iterator pa = ml.begin();
    pa.next();
    TS_ASSERT(pa.get() == &b);

    // Make another iterator and erase it
    ptr_multi_list_base::iterator pb = pa;
    ml.erase(pb);
    TS_ASSERT(pb.get() == 0);

    // Validate pa
    TS_ASSERT(pa.get() == 0);

    // Insert d
    ml.insert(pa, &d);

    // Validate iterator
    TS_ASSERT(pa.get() == 0);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &c);
    pa.next();
    TS_ASSERT(pa == ml.end());

    // Validate list
    pa = ml.begin();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &a);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &d);
    pa.next();
    TS_ASSERT(pa != ml.end());
    TS_ASSERT(pa.get() == &c);
    pa.next();
    TS_ASSERT(pa == ml.end());
}

/** Test typed operation. */
void
TestPtrMultilist::testTyped()
{
    TS_ASSERT_EQUALS(Obj::live, 0);

    {
        // Create something
        ptr_multi_list<Obj> ml;
        TS_ASSERT(ml.empty());
        TS_ASSERT_EQUALS(ml.size(), 0);
        ml.push_back(new Obj(1));
        TS_ASSERT(!ml.empty());
        TS_ASSERT_EQUALS(ml.size(), 1);
        ml.push_back(new Obj(2));
        ml.push_back(new Obj(3));
        TS_ASSERT(!ml.empty());
        TS_ASSERT_EQUALS(Obj::live, 3);
        TS_ASSERT_EQUALS(ml.size(), 3);

        // Regular iteration
        {
            ptr_multi_list<Obj>::iterator i = ml.begin();
            TS_ASSERT(i != ml.end());
            TS_ASSERT(i == ml.begin());
            TS_ASSERT_EQUALS((*i)->n, 1);
            ++i;
            TS_ASSERT(i != ml.end());
            TS_ASSERT(i != ml.begin());
            TS_ASSERT_EQUALS((*i)->n, 2);
            ++i;
            TS_ASSERT(i != ml.end());
            TS_ASSERT(i != ml.begin());
            TS_ASSERT_EQUALS((*i)->n, 3);
            ++i;
            TS_ASSERT(i == ml.end());
            TS_ASSERT(i != ml.begin());
        }

        // Regular iteration with two iterators
        {
            ptr_multi_list<Obj>::iterator i = ml.begin();
            ptr_multi_list<Obj>::iterator j = ml.begin();
            TS_ASSERT(i == j);
            TS_ASSERT_EQUALS((*i)->n, 1);
            TS_ASSERT_EQUALS((*j)->n, 1);
            ++i;
            TS_ASSERT(i != j);
            TS_ASSERT_EQUALS((*i)->n, 2);
            TS_ASSERT_EQUALS((*j)->n, 1);
            ++j;
            TS_ASSERT(i == j);
            TS_ASSERT_EQUALS((*i)->n, 2);
            TS_ASSERT_EQUALS((*j)->n, 2);
            TS_ASSERT_EQUALS(Obj::live, 3);
        }
        
        // Deletion of second element with both iterators pointing there
        {
            ptr_multi_list<Obj>::iterator i = ml.begin();
            ptr_multi_list<Obj>::iterator j = ml.begin();
            ++i;
            ++j;
            ml.erase(i);
            TS_ASSERT(*i == 0);
            TS_ASSERT(*j == 0);
            ++i;
            ++j;
            TS_ASSERT_EQUALS((*i)->n, 3);
            TS_ASSERT_EQUALS((*j)->n, 3);
            TS_ASSERT(i == j);
            TS_ASSERT_EQUALS(Obj::live, 2);
        }

        // List now has two elements. Try deleting the first.
        {
            ptr_multi_list<Obj>::iterator i = ml.begin();
            ptr_multi_list<Obj>::iterator j = ml.begin();
            ml.erase(i);
            TS_ASSERT(*i == 0);
            TS_ASSERT(*j == 0);
            ++i;
            ++j;
            TS_ASSERT_EQUALS((*i)->n, 3);
            TS_ASSERT_EQUALS((*j)->n, 3);
            TS_ASSERT_EQUALS(Obj::live, 1);
        }

        // List now has one element. Must be 3.
        {
            ptr_multi_list<Obj>::iterator i = ml.begin();
            TS_ASSERT_EQUALS((*i)->n, 3);
            TS_ASSERT_EQUALS(ml.size(), 1);
        }

        // Insert one element to test insert()
        {
            ml.insert(ml.begin(), new Obj(9));
            TS_ASSERT_EQUALS(Obj::live, 2);
            TS_ASSERT_EQUALS(ml.size(), 2);
            TS_ASSERT_EQUALS(ml.front()->n, 9);
        }
    }

    TS_ASSERT_EQUALS(Obj::live, 0);
}

/** Test typed merge operation. */
void
TestPtrMultilist::testTypedMerge()
{
    ptr_multi_list<Obj> ml;
    ml.push_back(new Obj(1));
    ml.push_back(new Obj(3));
    ml.push_back(new Obj(5));
    ml.push_back(new Obj(7));
    ml.push_back(new Obj(8));
    ml.push_back(new Obj(9));

    ptr_multi_list<Obj> ml2;
    ml2.push_back(new Obj(2));
    ml2.push_back(new Obj(4));
    ml2.push_back(new Obj(6));
    ml2.push_back(new Obj(10));
    ml2.push_back(new Obj(11));
    ml2.push_back(new Obj(12));

    TS_ASSERT_EQUALS(Obj::live, 12);
    TS_ASSERT_EQUALS(ml.size(), 6);
    TS_ASSERT_EQUALS(ml2.size(), 6);

    // Merge
    ml.merge(ml2, objCompare);
    TS_ASSERT_EQUALS(Obj::live, 12);
    TS_ASSERT_EQUALS(ml.size(), 12);
    TS_ASSERT_EQUALS(ml2.size(), 0);

    // Verify
    ptr_multi_list<Obj>::iterator p = ml.begin();
    for (int i = 1; i <= 12; ++i) {
        TS_ASSERT(p != ml.end());
        TS_ASSERT(*p != 0);
        TS_ASSERT_EQUALS((*p)->n, i);
        ++p;
    }
    TS_ASSERT(p == ml.end());

    // Merge and verify again (pathological case I)
    ml.merge(ml2, objCompare);
    TS_ASSERT_EQUALS(Obj::live, 12);
    TS_ASSERT_EQUALS(ml.size(), 12);
    TS_ASSERT_EQUALS(ml2.size(), 0);

    // Verify
    p = ml.begin();
    for (int i = 1; i <= 12; ++i) {
        TS_ASSERT(p != ml.end());
        TS_ASSERT(*p != 0);
        TS_ASSERT_EQUALS((*p)->n, i);
        ++p;
    }
    TS_ASSERT(p == ml.end());

    // Merge and verify again (pathological case II)
    ml2.merge(ml, objCompare);
    TS_ASSERT_EQUALS(Obj::live, 12);
    TS_ASSERT_EQUALS(ml2.size(), 12);
    TS_ASSERT_EQUALS(ml.size(), 0);

    // Verify
    p = ml2.begin();
    for (int i = 1; i <= 12; ++i) {
        TS_ASSERT(p != ml2.end());
        TS_ASSERT(*p != 0);
        TS_ASSERT_EQUALS((*p)->n, i);
        ++p;
    }
    TS_ASSERT(p == ml2.end());
}

/** Test typed sort operation. */
void
TestPtrMultilist::testTypedSort()
{
    ptr_multi_list<Obj> ml;
    ml.push_back(new Obj(6));
    ml.push_back(new Obj(9));
    ml.push_back(new Obj(1));
    ml.push_back(new Obj(4));
    ml.push_back(new Obj(2));
    ml.push_back(new Obj(5));
    ml.push_back(new Obj(7));
    ml.push_back(new Obj(3));
    ml.push_back(new Obj(8));

    TS_ASSERT_EQUALS(Obj::live, 9);
    TS_ASSERT_EQUALS(ml.size(), 9);

    // Sort
    ml.sort(objCompare);
    
    // Verify
    ptr_multi_list<Obj>::iterator p = ml.begin();
    for (int i = 1; i <= 9; ++i) {
        TS_ASSERT(p != ml.end());
        TS_ASSERT(*p != 0);
        TS_ASSERT_EQUALS((*p)->n, i);
        ++p;
    }
    TS_ASSERT(p == ml.end());
}
