/**
  *  \file t_sigslot.cc
  *  \brief Testsuite for sigslot.h
  *
  *  Note that this assumes that slots are called in reverse order of
  *  registration, although sigslot.h does not state any such property.
  *
  *  (c) 2001-2016 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/sigslot.h"

#include "tests.h"
#include "cpluslib/format.h"
#include "cpluslib/string.h"

static string_t accum;

/***************************** Parameterless *****************************/

namespace {
    class Receiver {
     public:
        int i;
        SignalConnection conn;
        Receiver(int i, Signal0& sig)
            : i(i), conn(sig.add(this, &Receiver::handle))
            { }
        void handle()
            { accum += format("*Rec%d", i); }
    };

    void staticFun()
    {
        accum += "*stat";
    }
}

void
TestSigSlot::testFirst()
{
    // explicit signal connect/disconnect; disconnect first handler
    Signal0 sig;
    {
        Receiver rec(0, sig);
        sig.add(staticFun);

        accum.clear();
        sig.raise();
        TS_ASSERT_EQUALS(accum, "*stat*Rec0");
    }
    accum.clear();
    sig.raise();
    TS_ASSERT_EQUALS(accum, "*stat");
}

void
TestSigSlot::testSecond()
{
    // explicit signal connect/disconnect; disconnect second handler
    Signal0 sig;
    {
        sig.add(staticFun);
        Receiver rec(0, sig);

        accum.clear();
        sig.raise();
        TS_ASSERT_EQUALS(accum, "*Rec0*stat");
    }
    accum.clear();
    sig.raise();
    TS_ASSERT_EQUALS(accum, "*stat");
}

void
TestSigSlot::testSigDies()
{
    // signal dies before receiver
    Signal0* sig = new Signal0;
    Receiver rec(0, *sig);

    TS_ASSERT(rec.conn.isConnected());

    accum.clear();
    sig->raise();
    TS_ASSERT_EQUALS(accum, "*Rec0");

    delete sig;
    TS_ASSERT(!rec.conn.isConnected());
}

/***************************** With Argument *****************************/

namespace {
    class ReceiverArg {
     public:
        int i;
        SignalConnection conn;
        ReceiverArg(int i, Signal1<int>& sig)
            : i(i), conn(sig.add(this, &ReceiverArg::handle))
            { }
        void handle(int n)
            {
                accum += format("*Rec%d,%d", i, n);
            }
    };

    void staticFunArg(int n)
    {
        accum += format("*stat%d", n);
    }
}

void
TestSigSlot::testFirstArg()
{
    // explicit signal connect/disconnect; disconnect first handler; with arg
    Signal1<int> sig;
    {
        ReceiverArg rec(0, sig);
        sig.add(staticFunArg);

        accum.clear();
        sig.raise(99);
        TS_ASSERT_EQUALS(accum, "*stat99*Rec0,99");
    }
    accum.clear();
    sig.raise(88);
    TS_ASSERT_EQUALS(accum, "*stat88");
}

void
TestSigSlot::testSecondArg()
{
    // explicit signal connect/disconnect; disconnect second handler; with arg
    Signal1<int> sig;
    {
        // we use staticFun here to test invocation without args
        sig.add(staticFun);
        ReceiverArg rec(0, sig);

        accum.clear();
        sig.raise(99);
        TS_ASSERT_EQUALS(accum, "*Rec0,99*stat");
    }
    accum.clear();
    sig.raise(88);
    TS_ASSERT_EQUALS(accum, "*stat");
}

void
TestSigSlot::testSigDiesArg()
{
    // signal dies before receiver
    Signal1<int>* sig = new Signal1<int>;
    ReceiverArg rec(0, *sig);

    TS_ASSERT(rec.conn.isConnected());

    accum.clear();
    sig->raise(77);
    TS_ASSERT_EQUALS(accum, "*Rec0,77");

    delete sig;
    TS_ASSERT(!rec.conn.isConnected());
}

void
TestSigSlot::testMulti()
{
    // four identical handlers
    Signal0 sig;
    sig.add(staticFun);
    sig.add(staticFun);
    sig.add(staticFun);
    sig.add(staticFun);

    accum.clear();
    sig.raise();
    TS_ASSERT_EQUALS(accum, "*stat*stat*stat*stat");
}

void
TestSigSlot::testMultiRec()
{
    // four objects
    Signal0 sig;
    Receiver r1(0, sig), r2(1, sig), r3(2, sig), r4(3, sig);

    accum.clear();
    sig.raise();
    TS_ASSERT_EQUALS(accum, "*Rec3*Rec2*Rec1*Rec0");
}
